/*	$Id: gikTrackCodels.cc 2008/04/16 14:44:55 mallet $ */

/*
 * Copyright (c) 2007 LAAS/CNRS                       --  Mon Apr 16 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: gikTrackCodels.cc 2008/04/16 14:44:55 mallet $";

#include <err.h>
#include <portLib.h>

#include "server/gikHeader.h"
#include "netshring.h"
#include "genom-openhrp/genom-hrp2.h"

#include "PaletteGik.h"
#include "core/hppGikSolver.h"
#include "hppGikTools.h"


/* --- TrackSpeed ---------------------------------------------------- */

/** Move a give hand at the velocity exported in a poster
 *
 * Reports:     OK
 *              S_gik_NO_MOTION
 *              S_gik_CANNOT_PUSH
 */

/* local variables */
enum {
  GIK_CONSTRAINT_FOOT,
  GIK_CONSTRAINT_COM,
  GIK_CONSTRAINT_HAND,
//   GIK_CONSTRAINT_GAZE,
  GIK_CONSTRAINT_TOTAL
};

static ChppGikSolver *solver;
static vector<CjrlGikStateConstraint *> constraints(GIK_CONSTRAINT_TOTAL);

static POSTER_ID objectId, wristId;
static T3D *objectPose, *wristPose;

static vector<vector3d> *cbufferwr;
static vector<matrix4d> *cbufferwa;
static unsigned int cindex;

/* gikTrackSpeedStart  -  codel START of TrackSpeed
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackSpeedStart(GikTrackSpeed *trackSpeed, int *report)
{
  /* --- look for posters --- */

  if (posterFind(trackSpeed->objectName.name, &objectId) != OK) {
    *report = S_gik_NOT_FOUND;
    return ETHER;
  }
  objectPose = (T3D *)posterAddr(objectId);

  if (posterFind(trackSpeed->wristName.name, &wristId) != OK) {
    *report = S_gik_NOT_FOUND;
    return ETHER;
  }
  wristPose = (T3D *)posterAddr(wristId);


  /* --- read current configuration --- */

  if (gikGetConfigMain(report) != OK) return ETHER;

  matrix4d init = robot->waist()->currentTransformation();
  vector3d initw;
  /* XXX hardcoded 50ms */
  cbufferwa = new vector<matrix4d>((size_t)(trackSpeed->delay * 20), init);

  switch(trackSpeed->hand) {
    case GIK_LEFT_HAND:
      init = robot->leftWrist()->currentTransformation();
      initw = init * robot->leftHand()->centerInWristFrame();
      break;
    case GIK_RIGHT_HAND:
      init = robot->rightWrist()->currentTransformation();
      initw = init * robot->rightHand()->centerInWristFrame();
      break;
  }
  cbufferwr = new vector<vector3d>((size_t)(trackSpeed->delay * 20), initw);
  cindex = 0;

  /* --- create stability related constraints -------------- */

  solver = new ChppGikSolver(robot);

  /* the support foot */
  robot->clearFixedJoints();
  robot->addFixedJoint(robot->rightFoot());

  /* the non support foot */
  vector3d point;
  point[0] = point[1] = point[2] = 0.0;

  constraints[GIK_CONSTRAINT_FOOT] =
    paletteGik->gikFactory.createTransformationConstraint(
      *robot, *robot->leftFoot(), point,
      robot->leftFoot()->currentTransformation());

  /* CoM constraint */
  vector3d com = robot->positionCenterOfMass();

  constraints[GIK_CONSTRAINT_COM] =
    paletteGik->gikFactory.createComConstraint(*robot, com[0], com[1]);


  /* --- hand constraint ----------------------------------- */

  matrix4d p;

  switch(trackSpeed->hand) {
    case GIK_LEFT_HAND:
      p = robot->leftWrist()->currentTransformation();
      constraints[GIK_CONSTRAINT_HAND] =
	paletteGik->gikFactory.createPositionConstraint(
	  *robot, *robot->leftWrist(),
	  robot->leftHand()->centerInWristFrame(),
	  p * robot->leftHand()->centerInWristFrame());
      break;

    case GIK_RIGHT_HAND:
      p = robot->rightWrist()->currentTransformation();
      constraints[GIK_CONSTRAINT_HAND] =
	paletteGik->gikFactory.createPositionConstraint(
	  *robot, *robot->rightWrist(),
	  robot->rightHand()->centerInWristFrame(),
	  p * robot->rightHand()->centerInWristFrame());
      break;
  }


  /* --- gaze constraint ----------------------------------- */

#if 0
  constraints[GIK_CONSTRAINT_GAZE] =
    paletteGik->gikFactory.createGazeConstraint(
      *robot, robot->rightHand()->centerInWristFrame());
#endif


  /* --- Joints weights --- */

  vectorN weights;

  ChppGikTools::combineMasks(
    paletteGik->attStandingRobot->maskFactory()->wholeBodyMask(),
    paletteGik->attStandingRobot->maskFactory()->weightsDoubleSupport(), weights);
  solver->weights(weights);

  return EXEC;
}

/* gikTrackSpeedMain  -  codel EXEC of TrackSpeed
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackSpeedMain(GikTrackSpeed *trackSpeed, int *report)
{
  static vector3d goal;
  matrix4d wp;
  T3D object, wrist;

  switch(trackSpeed->hand) {
    case GIK_LEFT_HAND:
      wp = robot->leftWrist()->currentTransformation();
      (*cbufferwr)[cindex] = wp * robot->leftHand()->centerInWristFrame();
      break;

    case GIK_RIGHT_HAND:
      wp = robot->rightWrist()->currentTransformation();
      (*cbufferwr)[cindex] = wp * robot->rightHand()->centerInWristFrame();
      break;
  }
  (*cbufferwa)[cindex] = robot->waist()->currentTransformation();
  cindex++;
  if (cindex >= cbufferwa->size()) cindex = 0;

  /* read posters */
  if (posterTake(objectId, POSTER_READ) != OK) return EXEC;
  object = *objectPose;
  posterGive(objectId);

  if (posterTake(wristId, POSTER_READ) != OK) return EXEC;
  wrist = *wristPose;
  posterGive(wristId);

  if (!t3dOk(&object) || !t3dOk(&wrist)) goto motion;

  /* update goal */
  t3dConvertTo(T3D_MATRIX, &object);
  t3dConvertTo(T3D_MATRIX, &wrist);

  {
    matrix4d t = (*cbufferwa)[cindex];
    MAL_S4x4_MATRIX_ACCESS_I_J(t,0,3) = 0.;
    MAL_S4x4_MATRIX_ACCESS_I_J(t,1,3) = 0.;
    MAL_S4x4_MATRIX_ACCESS_I_J(t,2,3) = 0.;

    vector3d delta;
    delta[0] = object.matrix[3] - wrist.matrix[3];
    delta[1] = object.matrix[7] - wrist.matrix[7];
    delta[2] = object.matrix[11] - wrist.matrix[11];

    goal = (*cbufferwr)[cindex] + t * delta;
  }
  cout << goal;

motion:
  /* update hand constraint */
  double n;
  vector3d delta;
  vector3d p = dynamic_cast<CjrlGikPositionConstraint *>
    (constraints[GIK_CONSTRAINT_HAND])->worldTarget();

  delta = goal - p;

  n = sqrt(delta[0]*delta[0]+delta[1]*delta[1]+delta[2]*delta[2]);
  if (n < 0.02) return END;
  if (n > trackSpeed->speed)
    delta = delta * trackSpeed->speed / n;
  else
    return END;

  dynamic_cast<CjrlGikPositionConstraint *>
    (constraints[GIK_CONSTRAINT_HAND])->worldTarget(p + delta);

  //compute the support foot jacobian
  robot->rightFoot()->computeJacobianJointWrtConfig();

  //compute constraints jacobians and values
  for (unsigned int i = 0; i<constraints.size(); i++) {
    constraints[i]->computeValue();
    constraints[i]->computeJacobian();
  }

  //Account for joint limits: modifies the weights according to current configuration
  solver->accountForJointLimits();

  //1-step Solve (might not fulfill the desired constraints)
  bool ok = solver->gradientStep(constraints);
  if (!ok) { std::cout << "too far\n"; return END; }

  //-- Retrieve solution --//
  vectorN motionConfig = robot->currentConfiguration();
  vector3d dummy, zmpw;
  static double motionTimeRef = 0.;
  ghrp2_config_t cfg;
  size_t s;
  int i;

  motionTimeRef += 0.05;
  cfg.time = motionTimeRef;
  for (i = 0; i<GHRP2_MAXDOF; i++) {
    cfg.angles[i] = motionConfig[i+6];
  }

  paletteGik->attStandingRobot->updateDynamics(
    0.05, dummy, dummy, zmpw, dummy);

  cfg.zmp[0] = zmpw[0];
  cfg.zmp[1] = zmpw[1];
  cfg.zmp[2] = zmpw[2];

  cfg.waistRpy[0] = motionConfig[3];
  cfg.waistRpy[1] = motionConfig[4];
  cfg.waistRpy[2] = motionConfig[5];

  s = netshring_push(shtarget, &cfg, sizeof(cfg));
  if (s < 0 || s != sizeof(cfg)) {
    *report = S_gik_CANNOT_PUSH;
    return ETHER;
  }

  return EXEC;
}

/* gikTrackSpeedEnd  -  codel END of TrackSpeed
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackSpeedEnd(GikTrackSpeed *trackSpeed, int *report)
{
  //clean the mess
  delete solver;
  for (unsigned int i = 0; i<constraints.size();i++)
    delete constraints[i];
  return ETHER;
}



/* --- TrackHands ---------------------------------------------------- */

/** Move the hands at the position exported in a file
 *
 * Reports:     OK
 *              S_gik_NO_MOTION
 *              S_gik_CANNOT_PUSH
 */

static vector<CjrlGikStateConstraint *> hands_stk(4);
static FILE *trajectory;
static FILE *fang, *frpy, *fzmp;

/* gikTrackSpeedStart  -  codel START of TrackSpeed
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackHandsStart(GikTrackHands *trackHands, int *report)
{
  char full[1024];

  /* --- look for file --- */
  trajectory = fopen(trackHands->file.name, "r");
  if (!trajectory) {
    *report = S_gik_NOT_FOUND;
    return ETHER;
  }

  /* open pos */
  sprintf(full, "%s.pos", trackHands->dump.name);
  fang = fopen(full, "w");
  if (!fang) {
    *report = S_gik_NOT_FOUND; return ETHER;
  }

  /* open rpy */
  sprintf(full, "%s.rpy", trackHands->dump.name);
  frpy = fopen(full, "w");
  if (!frpy) {
    fclose(fang);
    *report = S_gik_NOT_FOUND; return ETHER;
  }

  /* open zmp */
  sprintf(full, "%s.zmp", trackHands->dump.name);
  fzmp = fopen(full, "w");
  if (!fzmp) {
    fclose(fang); fclose(frpy);
    *report = S_gik_NOT_FOUND; return ETHER;
  }

  /* --- read current configuration --- */

  if (gikGetConfigMain(report) != OK) return ETHER;


  /* --- create stability related constraints -------------- */

  solver = new ChppGikSolver(robot);

  /* the support foot */
  robot->clearFixedJoints();
  robot->addFixedJoint(robot->rightFoot());

  /* the non support foot */
  vector3d point;
  point[0] = point[1] = point[2] = 0.0;

  hands_stk[0] =
    paletteGik->gikFactory.createTransformationConstraint(
      *robot, *robot->leftFoot(), point,
      robot->leftFoot()->currentTransformation());

  /* CoM constraint */
  vector3d com = robot->positionCenterOfMass();

  hands_stk[1] =
    paletteGik->gikFactory.createComConstraint(*robot, com[0], com[1]);


  /* --- hands constraint ----------------------------------- */

  matrix4d p;

  p = robot->leftWrist()->currentTransformation();
  hands_stk[2] =
    paletteGik->gikFactory.createPositionConstraint(
      *robot, *robot->leftWrist(),
      robot->leftHand()->centerInWristFrame(),
      p * robot->leftHand()->centerInWristFrame());

  p = robot->rightWrist()->currentTransformation();
  hands_stk[3] =
    paletteGik->gikFactory.createPositionConstraint(
      *robot, *robot->rightWrist(),
      robot->rightHand()->centerInWristFrame(),
      p * robot->rightHand()->centerInWristFrame());


  /* --- Joints weights --- */

  vectorN weights;

  ChppGikTools::combineMasks(
    paletteGik->attStandingRobot->maskFactory()->wholeBodyMask(),
    paletteGik->attStandingRobot->maskFactory()->weightsDoubleSupport(), weights);
  solver->weights(weights);

  return EXEC;
}

/* gikTrackHandsMain  -  codel EXEC of TrackHands
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackHandsMain(GikTrackHands *trackHands, int *report)
{
  static vector3d left, right;
  matrix4d wp;

  if (fscanf(trajectory, "%lf %lf %lf %lf %lf %lf",
	     &left[0], &left[1], &left[2],
	     &right[0], &right[1], &right[2]) != 6) {
    warnx("bad line in trajectory");
    return END;
  }
  printf("Left: %g %g %g -- Right: %g %g %g\n",
	left[0], left[1], left[2], right[0], right[1], right[2]);

  /* update hand constraint */
  dynamic_cast<CjrlGikPositionConstraint *>(hands_stk[2])
    ->worldTarget(left);
  dynamic_cast<CjrlGikPositionConstraint *>(hands_stk[3])
    ->worldTarget(right);


  //compute the support foot jacobian
  robot->rightFoot()->computeJacobianJointWrtConfig();

  //compute constraints jacobians and values
  for (unsigned int i = 0; i<hands_stk.size(); i++) {
    hands_stk[i]->computeValue();
    hands_stk[i]->computeJacobian();
  }

  //Account for joint limits: modifies the weights according to current configuration
  solver->accountForJointLimits();

  //1-step Solve (might not fulfill the desired constraints)
  bool ok = solver->gradientStep(hands_stk);
  if (!ok) { std::cout << "too far\n"; return END; }

  //-- Retrieve solution --//
  vectorN motionConfig = robot->currentConfiguration();
  vector3d dummy, zmpw;
  static double motionTimeRef = 0.;
  ghrp2_config_t cfg;
  size_t s;
  int i;

  motionTimeRef += trackHands->period;
  cfg.time = motionTimeRef;
  for (i = 0; i<GHRP2_MAXDOF; i++) {
    cfg.angles[i] = motionConfig[i+6];
  }

  paletteGik->attStandingRobot->updateDynamics(
    trackHands->period, dummy, dummy, zmpw, dummy);

  cfg.zmp[0] = zmpw[0];
  cfg.zmp[1] = zmpw[1];
  cfg.zmp[2] = zmpw[2];

  cfg.waistRpy[0] = motionConfig[3];
  cfg.waistRpy[1] = motionConfig[4];
  cfg.waistRpy[2] = motionConfig[5];

  s = netshring_push(shtarget, &cfg, sizeof(cfg));
  if (s < 0 || s != sizeof(cfg)) {
    *report = S_gik_CANNOT_PUSH;
    return ETHER;
  }

  fprintf(fang, "%g ", cfg.time);
  fprintf(frpy, "%g ", cfg.time);
  fprintf(fzmp, "%g ", cfg.time);
  for (i = 0; i<GHRP2_MAXDOF; i++) {
    fprintf(fang, "%g ", cfg.angles[i]);
  }
  fprintf(fang, "\n");

  fprintf(frpy, "%g %g %g\n",
	  cfg.waistRpy[0], cfg.waistRpy[1], cfg.waistRpy[2]);

  fprintf(fzmp, "%g %g %g\n",
	  cfg.zmp[0], cfg.zmp[0], cfg.zmp[0]);

  return EXEC;
}

/* gikTrackHandsEnd  -  codel END of TrackHands
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikTrackHandsEnd(GikTrackHands *trackHands, int *report)
{
  //clean the mess
  if (fang) fclose(fang);
  if (frpy) fclose(frpy);
  if (fzmp) fclose(fzmp);

  if (trajectory) fclose(trajectory);
  delete solver;
  for (unsigned int i = 0; i<hands_stk.size();i++)
    delete hands_stk[i];
  return ETHER;
}
